from .device import Device
from .battery_converter_device import BatteryConverterDevice
from .energy_manager_device import EnergyManagerDevice
from .location_device import LocationDevice
from .ev_station_device import EVStationDevice
from .s0_counter_device import S0CounterDevice
from .energy_manager_data import EnergyManagerData
from .energy_manager import EnergyManager



"""
Contains the definition of the EnergyManager devices and the parsing logic.

Not all devices here are modelled. This should be expanded as needed.
This has not been tested with MyReserve battery.

The following assumptions are made, and may need to be adjusted:
- Only one location device exists
- Only one energy manager device exists
- Only one battery may exist - needs to be extended for more batteries


The JSON data has the structure:
result: {
    items: [{
        tagValues: {
            tagName: string,
            guid: string,
            value: Any
        },
        deviceModel: [{
            deviceClass: string
        }, {
            deviceClass: string
        }],
        guid: string
    }]
}
"""

from typing import Dict, List, Optional
from . import BatteryConverterDevice
from . import Device
from . import EnergyManagerDevice
from . import LocationDevice
from . import EVStationDevice
from . import S0CounterDevice


class EnergyManagerData:
    """Contains all of the energy manager devices and the getter methods."""

    devices: Dict[str, Device] = {}

    def __init__(self, json: dict):
        self.load_json(json)

    def load_json(self, json: dict) -> None:
        """Read the json data from the EnergyManager."""
        if "result" in json:
            result = json["result"]
            if "items" in result:
                items = result["items"]
                for item in items:
                    device = Device()
                    device.read_json_item(item)
                    self.devices[device.guid] = device

    def get_devices_by_class(self, device_class: str) -> Optional[List[Device]]:
        """Get the devices with the specified class."""
        return (d for d in self.devices.values() if device_class in d.device_classes)

    def get_device_by_class(self, device_class: str) -> Optional[Device]:
        """Get the first instance of the device with the specified class."""
        return next(
            (d for d in self.devices.values() if device_class in d.device_classes), None
        )
    
    def get_device_by_guid(self, guid: str) -> Optional[Device]:
        """Get the device by GUID."""
        return self.devices.get(guid)

    @property
    def energy_manager_device(self) -> Optional[EnergyManagerDevice]:
        """Get the EnergyManagerDevice. Assumption that there is only one EnergyManager device."""
        device = self.get_device_by_class(EnergyManagerDevice.DEVICE_CLASS)
        return EnergyManagerDevice(device) if device else None

    @property
    def location_device(self) -> Optional[LocationDevice]:
        """Get the LocationDevice. Assumption that there is only one location device."""
        device = self.get_device_by_class(LocationDevice.DEVICE_CLASS)
        return LocationDevice(device) if device else None

    @property
    def battery_converter_devices(self) -> Optional[List[BatteryConverterDevice]]:
        """Get the BatteryConverterDevice objects."""
        devices = self.get_devices_by_class(BatteryConverterDevice.DEVICE_CLASS)
        return list(map(lambda device: BatteryConverterDevice(device), devices))

    @property
    def s0_counter_devices(self) -> Optional[List[S0CounterDevice]]:
        """Get the S0CounterDevice objects."""
        devices = self.get_devices_by_class(S0CounterDevice.DEVICE_CLASS)
        return list(map(lambda device: S0CounterDevice(device), devices))

    @property
    def ev_station_devices(self) -> Optional[List[EVStationDevice]]:
        """Get the EVStationDevice objects."""
        devices = self.get_devices_by_class(EVStationDevice.DEVICE_CLASS)
        return list(map(lambda device: EVStationDevice(device), devices))



from typing import Final
from . import Device

class EVStationDevice:
    """
    The electric vehicle charging station.
    """

    DEVICE_CLASS: Final[str] = "com.kiwigrid.devices.evstation.EVStation"

    # Tag name constants
    TAG_POWER_AC_IN: Final[str] = "PowerACIn"
    TAG_POWER_AC_OUT: Final[str] = "PowerACOut"
    TAG_WORK_AC_IN: Final[str] = "WorkACIn"
    TAG_WORK_AC_OUT: Final[str] = "WorkACOut"
    TAG_STATE_OF_CHARGE: Final[str] = "StateOfCharge"
    TAG_MODE_STATION: Final[str] = "ModeStation"
    TAG_WORK_AC_IN_SESSION: Final[str] = "WorkACInSession"
    TAG_TEMPERATURE_BATTERY: Final[str] = "TemperatureBattery"

    def __init__(self, device: Device):
        """Wrapper for EVStation."""
        self.device = device

    @property
    def power_ac_in(self) -> float:
        """The power flowing into the vehicle."""
        return self.device.get_tag_value_as_float(EVStationDevice.TAG_POWER_AC_IN)

    @property
    def power_ac_out(self) -> float:
        """The power flowing out of the vehicle."""
        return self.device.get_tag_value_as_float(EVStationDevice.TAG_POWER_AC_OUT)

    @property
    def work_ac_in(self) -> float:
        """The energy sent to the vehicle."""
        return self.device.get_tag_value_as_float(EVStationDevice.TAG_WORK_AC_IN)

    @property
    def work_ac_out(self) -> float:
        """The energy pulled from the vehicle."""
        return self.device.get_tag_value_as_float(EVStationDevice.TAG_WORK_AC_OUT)

    @property
    def state_of_charge(self) -> float:
        """The vehicle state of charge."""
        return self.device.get_tag_value_as_float(EVStationDevice.TAG_STATE_OF_CHARGE)

    @property
    def mode_station(self) -> str:
        """The current mode of the station."""
        return self.device.get_tag_value_as_str(EVStationDevice.TAG_MODE_STATION)

    @property
    def work_ac_in_session(self) -> float:
        """The energy sent to the vehicle for the session."""
        return self.device.get_tag_value_as_float(EVStationDevice.TAG_WORK_AC_IN_SESSION)

    @property
    def temperature_battery(self) -> float:
        """The battery temperature."""
        return self.device.get_tag_value_as_float(EVStationDevice.TAG_TEMPERATURE_BATTERY)



from typing import Final
from . import Device

class S0CounterDevice:
    """
    The S0 Counter device. The EnergyManager has two inputs available for S0 energy meters.
    """

    DEVICE_CLASS: Final[str] = "com.kiwigrid.devices.s0counter.S0Counter"

    # Tag name constants
    TAG_POWER_IN: Final[str] = "PowerIn"
    TAG_POWER_OUT: Final[str] = "PowerOut"
    TAG_WORK_IN: Final[str] = "WorkIn"
    TAG_WORK_OUT: Final[str] = "WorkOut"
    TAG_COUNT_PULSES: Final[str] = "CountPulses"

    def __init__(self, device: Device):
        """Wrapper for S0Counter."""
        self.device = device

    @property
    def power_in(self) -> float:
        """The power coming in, in W."""
        return self.device.get_tag_value_as_float(S0CounterDevice.TAG_POWER_IN)

    @property
    def power_out(self) -> float:
        """The power going out, in W."""
        return self.device.get_tag_value_as_float(S0CounterDevice.TAG_POWER_OUT)

    @property
    def work_in(self) -> float:
        """The energy coming in, in Wh."""
        return self.device.get_tag_value_as_float(S0CounterDevice.TAG_WORK_IN)

    @property
    def work_out(self) -> float:
        """The energy going out, in Wh."""
        return self.device.get_tag_value_as_float(S0CounterDevice.TAG_WORK_OUT)

    @property
    def count_pulses(self) -> int:
        """The pulse counter."""
        return self.device.get_tag_value_as_int(S0CounterDevice.TAG_COUNT_PULSES)
